/*
   Do not modify this file.
*/

import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer_Bounded extends Buffer
{
   private final Lock lock;
   private final Condition emptyCondition;
   private boolean empty;
   private final Condition fullCondition;
   private boolean full;

   /**
      Create a half filled buffer.

      @param c1  the left bracket for the string representation of the buffer
      @param c4  the right bracket for the string representation of the buffer
      @param SIZE  the size of the buffer
   */
   Buffer_Bounded(char c1, char c4, final int SIZE)
   {
      this(SIZE/2, c1, '0', '_', c4, SIZE);
   }


   /**
      Initialize a partially filled buffer.

      @param currentIndex  initially fill the buffer up to this index
      @param c1  the left bracket for the string representation of the buffer
      @param c4  the right bracket for the string representation of the buffer
      @param SIZE  the size of the buffer
   */
   Buffer_Bounded(int currentIndex, char c1, char c4, final int SIZE)
   {
      this(currentIndex, c1, '0', '_', c4, 80);
   }


   /**
      Create a partially filled buffer with a mutual exclusion lock,
      a condition variable that prevents buffer underflow, and
      a condition variable that prevents buffer overflow.

      @param currentIndex  initially fill the buffer up to this index
      @param c1  the left bracket for the string representation of the buffer
      @param c2  the char for the filled part of the buffer
      @param c3  the char for the empty part of the buffer
      @param c4  the right bracket for the string representation of the buffer
      @param SIZE  the size of the buffer
   */
   Buffer_Bounded(int currentIndex, char c1, char c2, char c3, char c4, final int SIZE)
   {
      super(currentIndex, c1, c2, c3, c4, SIZE); //Initialize a partially filled buffer.

      this.lock = new ReentrantLock();
      this.fullCondition  = lock.newCondition();
      this.emptyCondition = lock.newCondition();
      this.empty = (currentIndex == 0);
      this.full  = (currentIndex == SIZE);
   }


   /**
      Insert the given character at the next available buffer location.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to put in the buffer
   */
   @Override
   public void put(final char c)
   {
      try
      {
         this.lock.lock(); // mutual exclusion
         try
         {
            while (this.full)
            {
               this.fullCondition.await();
            }
            ++currentIndex;
            //Thread.yield(); // force race conditions to happen more often
            buf[currentIndex] = c;
            this.empty = (currentIndex == 0);
            this.full = (currentIndex == SIZE);
            this.emptyCondition.signal(); // can use signalAll OR signal()
         }
         finally
         {
            System.out.println(this);
            this.lock.unlock();
         }
      }
      catch (InterruptedException e)
      {
         Thread.currentThread().interrupt();
      }
   }


   /**
      Remove from the buffer the character at the buffer's current index.
      Mark the newly emptied spot in the buffer with the given character.
      Return the character that was removed from the buffer.

      Every time we do an operation on this buffer, it prints out its state.

      @param c  character to use to mark the newly emptied position in the buffer
      @return the character removed from the buffer
   */
   @Override
   public char get(final char c)
   {
      try
      {
         this.lock.lock(); // mutual exclusion
         try
         {
            while (this.empty)
            {
               this.emptyCondition.await();
            }
            final char result = buf[currentIndex];
            buf[currentIndex] = c;
            //Thread.yield(); // force race conditions to happen more often
            --currentIndex;
            this.empty = (currentIndex == 0);
            this.full = (currentIndex == SIZE);
            this.fullCondition.signal(); // can use signalAll OR signal()
            return result;
         }
         finally
         {
            System.out.println(this);
            this.lock.unlock();
         }
      }
      catch (InterruptedException e)
      {
         Thread.currentThread().interrupt();
      }
      return '\0'; // keep the compiler from complaining
   }
}
